#! /usr/local/bin/perl
#/*
#//  The Natto View
#//
#//  Copyright (c) 1995-1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
#//                           Keio University (Matsushita Lab.)
#//
#//  Please first read the full copyright statement in the file LICENSE.
#//  The code in this file may be used, copied and redistributed under the
#//  same terms as perl itself (see the files README, Copying and Artistic
#//  in the perl distribution).
#*/

use strict;
#use diagnostics;
#use sigtrap;
use vars qw($sock $db $nlinks $nurls %seen %url_number $http_from $log);

###############################################################################
#  SETTINGS

my $verbose    = 0;                          # output debugging messages?
my $link_port  = 33337;                      # service port
my $db_file    = '/home/WWW/etc/linkman.db'; # link cache database
my $log_file   = '/home/WWW/etc/natto.log';  # log file
my $def_user   = 'nattoview';                # default user for http header
my $def_domain = 'myo.inst.keio.ac.jp';      # default domain for http heaer
my $def_expire = 3600;                       # expiring period of link cache

my $http_timeout    = 60;  # timeout for http access
my $ident_timeout   = 10;  # timeout for identity check
my $request_timeout = 10;  # timeout for waiting request

# proxy settings 
my %proxies = {
#     'http:'  => 'http://www.myo.inst.keio.ac.jp:8080/',
#     'ftp:'   => 'http://www.myo.inst.keio.ac.jp:8080/',
      'dummy:' => 'http://www.myo.inst.keio.ac.jp:8080/'
};

###############################################################################

use Fcntl;
use Socket;
use FileHandle;
use Sys::Hostname;
use LWP::UserAgent;
use LWP::Socket;
use LWP::MediaTypes;
use URI::URL;
use HTTP::Status;
use HTML::Entities;
use HTML::Element;
use KanjiCode;
use LinkDB;
use Ident;
use LinkParser;

#use LWP::Debug;
#LWP::Debug::level('+');

my $hostname   = hostname;
my $version    = '0.5';
my $libwww     = '5.04';
my $user_agent = "NattoView LinkDB2/$version libwww-perl/$libwww";
my $def_depth  = 2;
my $max_depth  = 5;
my $max_listen = 10;
my $endl       = "\015\012";


sub main;
  sub connected;
    sub scan;
      sub get_node;

sub last_word;
sub dburl;
sub url_no;
sub sig_pipe;
sub sig_int;
sub sig_term;


# start as deamon
$SIG{'CHLD'} = sub { wait };
fork && exit;
main;
exit;


sub main
{
    $log = new FileHandle ">>$log_file";
    unless ($log) {
        print STDERR "Can't open log file $log_file.\n";
	exit;
    }
    
    my $listen = new LWP::Socket;
    setsockopt($listen->{'socket'}, SOL_SOCKET, SO_REUSEADDR, pack("l", 1));
    $listen->bind($hostname, $link_port);
    $listen->listen($max_listen);
    print STDERR "$hostname $user_agent - port=$link_port\n";

    $SIG{'INT'}  = sub { write_log('TERMINATE', "reason=sigint" );  exit };
    $SIG{'TERM'} = sub { write_log('TERMINATE', "reason=sigterm");  exit };
    write_log('START', "host=$hostname port=$link_port");

    $db = new LinkDB $db_file;

    for (;;) {
        my $pid;
	$sock = $listen->accept(10) || redo;
	if ($pid = fork) {
            undef $sock;
	}
	elsif (defined $pid) {
            undef $listen;
	    connected;
	    last;
        }
	elsif ($! =~ /No more process/) {
            sleep 5;
	    redo;
	}
    }
    exit;
}


sub connected
{
    my($rid, $host) = rident($sock->{'socket'}, $ident_timeout);
    $rid  = $def_user   unless $rid;
    $host = $def_domain unless $host;
    $http_from = "$rid\@$host";

    write_log('CONNECT', "ident=$http_from");
    $sock->write("200 L $version - $hostname $user_agent$endl");

    for (;;) {
        undef %seen;
	undef %url_number;
	$nurls = $nlinks = 0;

        $SIG{'PIPE'} = sub { write_log('SHUTDOWN', "reason=sigpipe");  exit };
        $SIG{'ALRM'} = sub { write_log('SHUTDOWN', "reason=timeout");  exit };

        alarm($request_timeout);
        $sock->read_until("\015?\012", \$_);
	alarm(0);
	write_log('request', "command=\"$_\"");
	
	my($cmd, $url, $depth) = split;
	if ($cmd =~ /^SCAN/i) {
	    $depth = ($depth && $depth =~ /^\d/) ? $depth: $def_depth;
	    if ($depth <= $max_depth) {
                $sock->write("210 L $version - OK.$endl" .
		             "0.0 NATTO $version NET$endl");
		scan(url($url)->abs->as_string, $depth);
		$sock->write("0.0 END$endl");
		write_log('response',"status=210 nurls=$nurls nlinks=$nlinks");
	    }
	    else {
                $sock->write("405 L $version - Parameter too large.$endl");
		write_log('response', "status=405");
	    }
	}
	elsif ($cmd =~ /^QUIT/i) {
            $sock->write("200 L $version - Connection closing.$endl");
	    write_log('response', "status=200");
	    last;
	}
	else {
            $sock->write("400 L $version - Unknown request.$endl");
	    write_log('response', "status=400");
	}
    }
    undef $sock;
    undef $db;
    write_log('SHUTDOWN', "reason=command");
}


sub scan
{
    my($from, $depth, $referer) = @_;
    return undef if $depth <= 0 || !$from;
    $from =~ s/#.*$//;
    
    my($url, $record)
        = get_node(url($from), ($depth <= 1) ? 'HEAD': 'GET', $referer);
    return unless $url;
    
    my $ctype = $$record{'CT'} ? $$record{'CT'}: 'error/000';
    my $base  = $$record{'BA'} ? url($$record{'BA'}, $url)->abs: $url->abs;
    my $title = $$record{'TI'} ? $$record{'TI'}: last_word($url);
    my $last_modified = $$record{'LM'} ? $$record{'LM'}: 0;
    my $expires       = $$record{'EX'} ? $$record{'EX'}: 0;
    my $last_get      = $$record{'GA'} ? $$record{'GA'}: 0;
    my $last_head     = $$record{'HA'} ? $$record{'HA'}: 0;
    my $i = url_no(url($from));
    $sock->write("0.0 URL $i $ctype $url $title$endl" .
                 "0.5 TIME $last_modified $expires $last_get $last_head$endl");
    return if $depth <= 1;

    my $n;
    foreach $n (1 .. $$record{'NL'}) {
        my $link   = $$record{"L$n"};
	my $anchor = $$record{"A$n"};
        my $to = url($link, $base)->abs;
	scan($to, $depth - 1, $url);
	my $frag = $to->frag ? $to->frag: '' ;
	$nlinks++;
	my $j = url_no($to);
	$sock->write("0.0 LINK $nlinks $i $j #$frag $anchor$endl");
    }
    1;
}


sub get_node
{
    my($url, $method, $referer) = @_;
    $method = ($method =~ /HEAD/i) ? 'HEAD': 'GET';
    my $record;
    
    print STDERR "$method $url: local..." if $verbose;
    for (;;) {
        if (!$url || $seen{"$url"} && $seen{"$url"} =~ /$method/i) {
            print STDERR "seen\n" if $verbose;
	    return (undef, undef);
	}
	$record = $db->get($url);
	last unless $$record{'RE'};
	$url = url($$record{'RE'}, $url)->abs;
    }
    $seen{"$url"} = "HEAD$method";

    my $now = time;
    if (($$record{'EX'} && $now < $$record{'EX'}) ||
        ($$record{'GA'} && $now < $$record{'GA'} + $def_expire) ||
	($$record{'HA'} && $now < $$record{'HA'} + $def_expire &&
	 $method eq 'HEAD')) {
        print STDERR "cached\n" if $verbose;
        return ($url, $record);
    }

    $sock->write("0.0 ECHO Request: $method $url...$endl");
    
    print STDERR "WWW" if $verbose;
    my $ua = new LWP::UserAgent $user_agent;
    my($scheme, $proxy);
    while (($scheme, $proxy) = each(%proxies)) {
        $ua->proxy($scheme, $proxy);
	print STDERR "(proxy)" if $verbose;
    }
    $ua->agent($user_agent);
    $ua->timeout($http_timeout);
    my $req = new HTTP::Request $method, $url;
    print STDERR "..." if $verbose;
    if ($method eq 'GET' && $$record{'GA'}) {
        $req->if_modified_since($$record{'GA'});
    }
    elsif ($$record{'HA'} && $$record{'GA'} &&
           $$record{'HA'} > $$record{'GA'}) {
        $req->if_modified_since($$record{'HA'});
    }
    $req->date($$record{'HA'} = time);
    $$record{'GA'} = $$record{'HA'} if $method eq 'GET';
    $req->referer("$referer") if $referer;
#    $req->from($http_from) if $http_from;

    my $res = $ua->request($req);
    my $status = $res->code;
    print STDERR "$status\n" if $verbose;

    if ($url ne $res->request->url) {
        $$record{'RE'} = $res->request->url->as_string;
        $db->set($url, $record);
	undef $$record{'RE'};
	$url = url($res->request->url, $url)->abs;
	$seen{"$url"} = "HEAD$method";
    }
    
    $$record{'CT'} = $res->content_type || guess_media_type($res->request->url);
    $$record{'LM'} = $res->last_modified;
    $$record{'EX'} = $res->expires;
    $$record{'TI'} = $res->title if $res->title;

    if ($status == RC_NOT_MODIFIED) {
        $db->set($url, $record);
        return ($url, $record);
    }
    if (is_error($status)) {
        $$record{'CT'} = "error/$status";
        $db->set($url, $record);
	return ($url, $record);
    }
    if ($$record{'CT'} !~ /html/) {
        $db->set($url, $record);
	return ($url, $record);
    }
    if ($method eq 'HEAD') {
        $db->set($url, $record);
	return ($url, $record);
    }

    any2euc($res->{'_content'});
    my $parser = new LinkParser $record;
    $parser->parse($res->content);
    $db->set($url, $record);
    ($url, $record);
}


sub last_word
{
    my $url = shift;
    my $str = "$url";
    $str =~ s|/$||;
    $str =~ s|^.*/||;
    $str =~ s|^.*:||;
    $str;
}


sub dburl
{
    my $url = shift->clone;
    my $path = $url->epath;
    $path =~ s|/$|| if $path ne '/';
    $url->epath($path);
    $url;
}


sub url_no
{
    my($url) = @_;
    $url = url($url) unless ref $url;
    $url = dburl($url->abs);
    my $urlstr = "$url";
    $urlstr =~ s/#.*$//;
    $url_number{$urlstr} = ++$nurls unless $url_number{$urlstr};
    $url_number{$urlstr};
}


sub write_log
{
    my($reason, $info) = @_;
    $reason = sprintf('%-9s', $reason);
    my $now = localtime;
    print $log "$$ $reason [$now] $info\n";
    print STDERR "$$ $reason [$now] $info\n" if $verbose;
}
