/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef WEB_H
#define WEB_H
#ifdef __GNUG__
#pragma interface
#endif

#include <istream>
#include <ostream>
#include <string>
#include "graph.h"
#include "dfs.h"
#include "ptrvector.h"
#include "observer.h"
#include "www.h"
#include "NodeInfo.h"

class WebExtender;
class WebReader;
class WebWriter;
class NodeMapper;


struct WebNotify
{
    enum priority_t { none, message, ignorable, sync };
    WebNotify(priority_t p, string m = ""): priority(p), msg(m) {}
    int    priority;
    string msg;
};


class Web: public observable<WebNotify>
{
 public:
    typedef graph<URL, NodeInfo, LinkInfo> graph_type;
    
    // types
    typedef graph_type::vertex node;
    typedef graph_type::edge   link;
    typedef graph_type::vertex_iterator node_iterator;
    typedef graph_type::edge_iterator   link_iterator;
    typedef graph_type::vertex_reverse_iterator node_reverse_iterator;
    typedef graph_type::edge_reverse_iterator   link_reverse_iterator;
    typedef ptrvector<node*> node_vector;
    typedef ptrvector<link*> link_vector;
#ifdef STRICT_STL
    typedef map<string, node*, less<string> > mark_map;
#else
    typedef map<string, void*, less<string> > mark_map;
#endif    

    Web(NodeMapper*);
    virtual ~Web();
    
    // graph management
    node* insert(const node& v) { return graph_->insert(v); }
    link* insert(const link& e) { return graph_->insert(e); }
    void  erase(node* v) { graph_->erase(v); }
    void  erase(link* e) { graph_->erase(e); }

    node* redirect(node*, URL);
    node* redirect(node*, node*);
    node* active_node() { return active_node_; }
    void  active_node(node*);
    link* active_link() { return active_link_; }
    void  active_link(link*);
    
    // lift-up and -down operations
    void lift_up(node*);
    void lift_up()   { lift_up(active_node_); }
    void lift_down(node*);
    void lift_down() { lift_down(active_node_); }
    void lift_down_all();

    // mark management
    mark_map& marks() { return marks_; }
    void      set_mark(string, node*);
    void      set_mark(string mark) { set_mark(mark, active_node_); }
    node*     get_mark(string);

    void update(WebNotify::priority_t p = WebNotify::sync, string msg = "")
	{ changed(true);  notify(WebNotify(p, msg)); }
    void message(string msg)
	{ changed(true);  notify(WebNotify(WebNotify::none, msg)); }

    graph_type::vertex_container& nodes() { return graph_->verteces(); }
    graph_type::edge_container&   links() { return graph_->edges(); }

    void clear_visited();

    // member access
    NodeMapper* mapper() const        { return mapper_; }
    void        mapper(NodeMapper* m) { mapper_ = m; }

    // void layout(node*);
    void layout_all();
				       
 private:
    graph_type* graph_;       // representaion
    node*       active_node_; // active node
    link*       active_link_; // active link
    mark_map    marks_;       // mark-onto-node map
    NodeMapper* mapper_;
};



class ArrangeNodes
{
    Web* web_;
    int  dir_;
    int  first_level_;
 public:
    enum direction_t { up, down };

    ArrangeNodes(Web* web, direction_t dir, int level):
	web_(web), dir_(dir), first_level_(level) {}
    int operator()(Web::node* vp, Web::link* ep, int depth);
};

#endif /* WEB_H */
