/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1996  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "WebIO.h"

#include <strstream>
#include "copy_line.h"
#include "NodeMapper.h"


WebReader::WebReader(
    Web*     web,
    istream* in,
    int      flags):
    WebIO(web, flags),
    first_node_(NULL),
    in_(in), ver_(1.0), type_("NET")
{}


WebReader::~WebReader()
{
    if (flags() & auto_free) delete in_;
}


void
WebReader::run()
{
    while (read_line())
	;
    web()->update(WebNotify::sync);
    delete this;
}


bool
WebReader::read_line()
{
    strstream buf;
    copy_line(*in_, buf); buf << ends;
    DEBUGOUT(buf.str());
    buf.freeze(false);

    float line_version = 0.0;
    buf >> ws >> line_version;
    if (line_version > 0.5) return true;

    string magic;
    buf >> ws >> magic;

    if (magic == "NATTO") {
	buf >> ver_ >> type_;
	strstream msg;
	msg << "NATTO " << ver_ << ' ' << type_ << ends;
	web()->message(msg.str());
	msg.freeze(false);
    }
    else if (magic == "URL") {
	unsigned n = 0;
	URL url;
	string mime;
	strstream title;
	buf >> n >> mime >> url;
	copy_line(buf, title); title << ends;
	if (n <= 0) return true;
	
	strstream msg;
	msg << "URL " << n << ' '
	    << mime << ' ' << url << ' ' << title.str() << ends;
	web()->message(msg.str());
	msg.freeze(false);
	
	Web::node* vp;
	Web::node v(url, NodeInfo(title.str(), mime));
	Web::node_iterator vi = web()->nodes().find(&v);
	if (vi == web()->nodes().end()) {
	    vp = web()->insert(v);
	}
	else {
	    vp = *vi;
	    vp->info().content_type = mime;
	    vp->info().title = title.str();
	}
	title.freeze(false);

	if (n == 1) {
	    if (first_node() && first_node() != vp)
		web()->redirect(first_node(), vp);
	    first_node_ = vp;
	    if (flags() | activate) web()->active_node(vp);
	}

	if (vvec_.size() > n) vvec_[n] = vp;
	else while (vvec_.size() <= n) vvec_.push_back(vp); 
    }
    else if (magic == "SAVE") {
	int n = vvec_.size() - 1;
	if (n <= 0) return true;
	NodeInfo& info = vvec_[n]->info();
	buf >> info.doi >> info.scanned;
    }
    else if (magic == "TIME") {
	int n = (int) vvec_.size() - 1;
	if (n <= 0) return true;
	time_t last_get, last_head;
	NodeInfo& info = vvec_[n]->info();
	buf >> info.last_modified >> info.expires >> last_get >> last_head;
	info.last_get = last_get;
    }
    else if (magic == "LINK") {
	unsigned a = 0, n1 = 0, n2 = 0;
	string frag;
	strstream anchor;
	buf >> a >> n1 >> n2 >> frag;
	copy_line(buf, anchor); anchor << ends;
	frag = frag.substr(1);
	
	strstream msg;
	msg << "LINK " << a << ' ' << n1 << ' ' << n2
	    << " #" << frag << ' ' << anchor.str() << ends;
	web()->message(msg.str());
	msg.freeze(false);
	anchor.freeze(false);

	if (n1 && n2 && n1 < vvec_.size() && n2 < vvec_.size() && n1 != n2) {
	    LinkInfo info(frag, anchor.str());
	    Web::link_iterator ei;
	    for (ei = vvec_[n1]->out().begin();
		 ei != vvec_[n1]->out().end() && (*ei)->info() != info; ei++)
		;
	    if (ei == vvec_[n1]->out().end()) {
		web()->insert(Web::link(vvec_[n1], vvec_[n2], info));
		if (vvec_[n2]->info().doi < vvec_[n1]->info().doi - 1)
		    vvec_[n2]->info().doi = vvec_[n1]->info().doi - 1;
		web()->mapper()->layout(vvec_[n1]);
		web()->mapper()->layout(vvec_[n2]);
		web()->update(WebNotify::ignorable);
	    }
	}
    }
    else if (magic == "ECHO") {
	strstream msg;
	copy_line(buf, msg); msg << ends;
	web()->message(msg.str());
	msg.freeze(false);
    }
    else if (magic == "COMM") {
	/* do nothing */;
    }
    else {
	return false;
    }
    return true;
}


WebWriter::WebWriter(
    Web*     web,
    ostream* out,
    int      flags):
    WebIO(web, flags), out_(out), type_("NET")
{}


WebWriter::~WebWriter()
{
    if (flags() & auto_free) delete out_;
}


void
WebWriter::run()
{
    *out_ << "0.0 NATTO 0.5 " << type_ << ' '
	<< web()->nodes().size() << ' ' << web()->links().size() << endl;
    
    int i = 1;
    Web::node_iterator vi = web()->nodes().begin();
    for (; vi != web()->nodes().end(); vi++, i++) {
	NodeInfo& info = (*vi)->info();
	string ct = info.content_type;
	if (!ct.length()) ct = "unknown";
	info.written = i;
	*out_ << "0.0 URL " << i << ' ' << ct << ' '
	    << (*vi)->key() << ' ' << info.title << endl
	    << "0.5 SAVE " << info.doi << ' ' << info.scanned << endl
	    << "0.5 TIME " << info.last_modified << ' ' << info.expires << ' '
	    << info.last_get << ' ' << info.last_head << endl
	    << "0.0 COMM 3D "
	    << info.x << ' ' << info.y << ' ' << info.z << ' '
	    << info.r << endl
	    << "0.0 COMM 2D " << info.x_2d << ' ' << info.y_2d << endl;
    }
    i = 1;
    Web::link_iterator ei = web()->links().begin();
    for (; ei != web()->links().end(); ei++, i++) {
	LinkInfo& info = (*ei)->info();
	*out_ << "0.0 LINK " << i << ' '
	    << (*ei)->from()->info().written << ' '
	    << (*ei)->to()->info().written << " #"
	    << info.frag << ' ' << info.anchor << endl;
    }
    *out_ << "0.0 END" << endl;
    delete this;
}

WebExtender::WebExtender(
    Web*       web,
    Web::node* first,
    int        depth,
    int        flags):
    WebIO(web, flags), first_node_(first), depth_(depth)
{}


WebExtender::WebExtender(
    Web* w,
    URL  url,
    int  depth,
    int  flags):
    WebIO(w, flags), first_node_(NULL), depth_(depth)
{
    Web::node v(url, NodeInfo(url.last_word()));
    Web::node_iterator vi = web()->nodes().find(&v);
    first_node_ = (vi != web()->nodes().end()) ? *vi: web()->insert(v);
}


WebExtender::~WebExtender()
{
    if (!first_node()) return;
    first_node()->info().scanned = true;
    first_node()->info().label_changed = true;
}
