/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifdef __GNUG__
#pragma implementation
#endif

#include "linkview.h"
#include "WebViewGL.h"

#include <math.h>
#include <stack>  /* #include <queue> */
#include <strstream>
#include <GL/glu.h>
#include <stdlib.h>
#include <string.h>
#include "glex.h"
#ifdef PABLO
#include "face.h"
#endif
#include "NodeMapper.h"

#define invalid_name ((GLuint)(-1))

#define DEF_PERS_FOVY      40.0
#define DEF_PERS_NEAR       0.3 
#define DEF_PERS_FAR       10.0 

static GLfloat light_ambient[]      = { 0.2, 0.2,  0.2, 1.0 };
static GLfloat light_diffuse[]      = { 0.8, 0.8,  0.8, 1.0 };
static GLfloat light_specular[]     = { 0.1, 0.1,  0.1, 1.0 };
static GLfloat light_position[]     = { 0.0, 0.0, 10.0, 0.0 };
static GLfloat light_shininess      = 10.0;

//static GLfloat material_ambient[]   = { 0.4, 0.4, 0.4, 1.0 };
static GLfloat material_diffuse[]   = { 0.8, 0.8, 0.8, 1.0 };
static GLfloat material_specular[]  = { 0.8, 0.8, 0.8, 1.0 };

#ifdef MESA
static GLfloat background0_color[]  = { 0.2, 0.2, 0.5, 0.0 };
#else
static GLfloat background0_color[]  = { 0.1, 0.1, 0.3, 0.0 };
#endif
static GLfloat background1_color[]  = { 0.0, 0.0, 0.0, 0.0 };
static GLfloat cursor_color[]       = { 0.0, 0.7, 1.0 };
static GLfloat board0_color[]       = { 0.5, 0.2, 0.7 };
static GLfloat board1_color[]       = { 0.2, 0.2, 0.5 };
static GLfloat board2_color[]       = { 0.6, 0.2, 0.2 };
static GLfloat board3_color[]       = { 0.7, 0.5, 0.1 };
static GLfloat board4_color[]       = { 0.5, 0.7, 0.2 };

static GLfloat normal_node_color[]  = { 1.0, 0.8, 0.0 };
static GLfloat scanned_node_color[] = { 0.8, 0.6, 0.0 };
static GLfloat active_node_color[]  = { 1.0, 0.4, 0.0 };
static GLfloat marked_node_color[]  = { 0.4, 0.7, 0.0 };

static GLfloat normal_link_color[]  = { 0.8, 0.7, 0.8, 0.3 };
//static GLfloat path_link_color[]    = { 0.8, 0.8, 0.8, 0.5 };
static GLfloat active_link_color[]  = { 0.4, 1.0, 0.0, 0.6 };

static GLfloat normal_text_color[]  = { 0.8, 0.8, 0.8 };
static GLfloat strong_text_color[]  = { 0.8, 1.0, 1.0 };
static GLfloat popup_text_color[]   = { 1.0, 1.0, 1.0 };
static GLfloat popup_plate_color[]  = { 0.3, 0.3, 0.6 };
static GLfloat active_plate_color[] = { 0.5, 0.3, 0.0 };
static GLfloat marked_plate_color[] = { 0.1, 0.4, 0.5 };

//static GLfloat status_text_color[]  = { 1.0, 1.0, 1.0 };

static GLfloat normal_link_width    = 1.0;
#ifdef MESA
static GLfloat active_link_width    = 1.5;
static GLfloat normal_text_width    = 1.5;
static GLfloat cursor_width         = 1.5;
#else
static GLfloat active_link_width    = 1.0;
static GLfloat normal_text_width    = 1.0;
static GLfloat cursor_width         = 1.0; 
#endif

//static GLint   fog_mode             = GL_EXP2;
//static GLfloat fog_density          = 0.4;
//static GLfloat fog_color[]          = { 0.0, 0.0, 0.0, 1.0 };


WebViewGL::WebViewGL(
    WebFrame* frame):
    fill_bitmap_(NULL), fill_bytes_(0),
    WebView(frame)
{}


#ifdef PABLO
WebViewGL::WebViewGL(
    WebViewGL* viewgl,
    int x,
    int y,
    int width,
    int height):
    fill_bitmap_(NULL), fill_bytes_(0),
    WebView((WebView*) viewgl, x, y, width, height)
{}
#endif


void
WebViewGL::init_graphics()
{
    make_current();
    
    glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);
    glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
    glHint(GL_FOG_HINT, GL_FASTEST);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
    
    glDepthFunc(GL_LEQUAL);
    glCullFace(GL_BACK);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glEnable(GL_LINE_SMOOTH);

    glLightfv(GL_LIGHT0, GL_AMBIENT,   light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,   light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR,  light_specular);
    glLightfv(GL_LIGHT0, GL_POSITION,  light_position);
    glLightf (GL_LIGHT0, GL_SHININESS, light_shininess);
    glEnable(GL_LIGHT0);

    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, material_specular);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#if 0    
    glFogi (GL_FOG_MODE,    fog_mode);
    glFogfv(GL_FOG_COLOR,   fog_color);
    glFogf (GL_FOG_DENSITY, fog_density);
#endif
    if (level() & 01) ClearColor(background1_color);
    else              ClearColor(background0_color);

    PackingMode pm;
    pm.swap_bytes = GL_FALSE;
    pm.lsb_first  = GL_FALSE;
    pm.row_length = pm.skip_rows = pm.skip_pixels = 0;
    pm.alignment  = 1;
    SetUnpackingMode(pm);
}


void
WebViewGL::resize(
    int x, int y,
    int w, int h)
{
    make_current();
    WebView::resize(x, y, w, h);
    glViewport(0, 0, w, h);
}


#ifndef PABLO
void
WebViewGL::projection_matrix()
{
    GLdouble fovy   = DEF_PERS_FOVY;
    GLdouble aspect = (GLdouble) width() / height();
    gluPerspective(fovy, aspect, DEF_PERS_NEAR, DEF_PERS_FAR);
}


void
WebViewGL::modelview_matrix()
{
    glTranslatef(trans().x, trans().y, trans().z);
    glRotatef(rot().x, 1.0, 0.0, 0.0);
    glRotatef(rot().y, 0.0, 1.0, 0.0);
    glRotatef(rot().z, 0.0, 0.0, 1.0);
    glTranslatef(-origin().x, -origin().y, -origin().z);
}
#endif


void
WebViewGL::update_prefix()
{
    make_current();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    if (level() & 01) ClearColor(background1_color);
    else              ClearColor(background0_color);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    projection_matrix();
}


void
WebViewGL::update_suffix()
{
    swap_buffers();
}


void
WebViewGL::draw_prefix()
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    modelview_matrix();

    glInitNames();
    glPushName(invalid_name);

    switch (render_mode()) {
     case wireframe:
	glShadeModel(GL_FLAT);
	break;
     case surface:
     case face_icon:
	glShadeModel(GL_SMOOTH);
	break;
    }
}


void
WebViewGL::draw_suffix()
{
    glPopName();
}


void
WebViewGL::draw_nodes_prefix()
{
    while (!node_table_.empty()) node_table_.pop_back();
}


void
WebViewGL::draw_a_node(
    Web::node* vp,
    int        status)
{
    int name = node_bit | node_table_.size();
    glLoadName(name);
    node_table_.push_back(vp);
    glPushMatrix();
    
    NodeInfo& info = vp->info();
    glTranslatef(info.x, info.y, info.z);
    glRotatef(180 - rootview()->rot().z, 0.0, 0.0, 1.0);
    glRotatef(      rootview()->rot().y, 0.0, 1.0, 0.0);
    glRotatef( 90 + rootview()->rot().x, 1.0, 0.0, 0.0);

#if 0
    if (info.icon_changed) {
	if (!info.icon_list) info.icon_list = glGenLists(1);
	glNewList(info.icon_list, GL_COMPILE_AND_EXECUTE);
	draw_a_node_icon(vp, status);
	glEndList();
	info.icon_changed = false;
    }
    else {
	glCallList(info.icon_list);
    }
#else
    if (info.icon_changed) {
	if (!info.icon_list) info.icon_list = glGenLists(1);
	glNewList(info.icon_list, GL_COMPILE);
	draw_a_node_icon(vp, status);
	glEndList();
	info.icon_changed = false;
    }
    glCallList(info.icon_list);
#endif

    glTranslatef(0, 0, info.r + 0.01);
#if 0
    if (info.label_changed) {
	if (!info.label_list) info.label_list = glGenLists(1);
	glNewList(info.label_list, GL_COMPILE_AND_EXECUTE);
	draw_a_node_label(vp, status);
	glEndList();
	info.label_changed = false;
    }
    else {
	glCallList(info.label_list);
    }
#else
    if (info.label_changed) {
	if (!info.label_list) info.label_list = glGenLists(1);
	glNewList(info.label_list, GL_COMPILE);
	draw_a_node_label(vp, status);
	glEndList();
	info.label_changed = false;
    }
    glCallList(info.label_list);
#endif
    
    glPopMatrix();
    glLoadName(invalid_name);
}


void
WebViewGL::draw_a_node_icon(
    Web::node* vp,
    int        status)
{
    NodeInfo& info = vp->info();
    
    if (render_mode() == wireframe) {
	FColor3fv((status & active_bit)  ? active_node_color:
		  (status & marked_bit)  ? marked_node_color:
		  (status & scanned_bit) ? scanned_node_color:
		  /* normal node */        normal_node_color, 0.9);
	// WireSphere(info.r, 8, 8);
	WireSphere_icosahedron(info.r, 1);
	return;
    }
    
    glEnable(GL_LIGHTING);
    Color3fv((status & active_bit)  ? active_node_color:
	     (status & marked_bit)  ? marked_node_color:
	     (status & scanned_bit) ? scanned_node_color:
	     /* normal node */        normal_node_color);
#ifdef PABLO
    if (render_mode() == face_icon)
	draw_face(info.r, info.eye, info.eyebrow, info.mouth);
    else
#endif
	SolidSphere_icosahedron(info.r, 1);

    glDisable(GL_LIGHTING);
}


void
WebViewGL::draw_a_node_label(
    Web::node* vp,
    int        status)
{
    NodeInfo& info = vp->info();
    string title =
	(status & active_bit) ? info.title: info.title.substr(0, 8);
    if (status & marked_bit) {
	title += "(";
	title += info.mark_keys.back().substr(0, 8);
	title += ")";
    }
    int width = 1, height = 1, base_line = 0;
    GLubyte* bitmap = string_bitmap(title, width, height, base_line);

    // plate
    int bytes = (width + 4 + 7) / 8 * (height + 2);
    if (bytes > fill_bytes_) {
	delete[] fill_bitmap_;
	fill_bytes_ = bytes;
	fill_bitmap_ = new GLubyte[fill_bytes_];
	memset(fill_bitmap_, 0xff, fill_bytes_);
    }
    Color3fv((status & active_bit) ? active_plate_color:
	     (status & marked_bit) ? marked_plate_color:
	     /* normal node */       popup_plate_color);
    glRasterPos3i(0, 0, 0);
    glBitmap(0, 0, 0, 0, -3 -2, -base_line -1, NULL);  // Mesa bug?
    glBitmap(width + 4, height + 2, 0, 0, 0, 0, fill_bitmap_);

    // label
    Color3fv(popup_text_color);
    glRasterPos3i(0, 0, 0);
    glBitmap(0, 0, 0, 0, -3, -base_line, NULL);  // Mesa bug?
    glBitmap(width, height, 0, 0, 0, 0, bitmap);
    delete[] bitmap;
}


void
WebViewGL::draw_nodes_suffix()
{}


void
WebViewGL::draw_links_prefix()
{
    glDepthMask(GL_FALSE);
    if (render_mode() >= surface) glEnable(GL_BLEND);
    while (!link_table_.empty()) link_table_.pop_back();
}

    
void
WebViewGL::draw_a_link(
    Web::link* ep,
    int        status)
{
    int name = link_bit | link_table_.size();
    glLoadName(name);
    link_table_.push_back(ep);

#if 0    
    LinkInfo& info = ep->info();
    if (info.changed) {
	if (!info.gl_list) info.gl_list = glGenLists(1);
	glNewList(info.gl_list, GL_COMPILE_AND_EXECUTE);
	draw_a_link_line(ep, status);
	glEndList();
    }
    else {
	glCallList(info.gl_list);
    }
#else
    draw_a_link_line(ep, status);
#endif    

    glLoadName(invalid_name);
}

		   
void
WebViewGL::draw_a_link_line(
    Web::link* ep,
    int        status)
{
    if (render_mode() >= surface)
	Color4fv((status & active_bit) ? active_link_color:
		 /* normal link */       normal_link_color);
    else
	FColor3fv((status & active_bit) ? active_link_color:
		  /* normal link */       normal_link_color, 0.6);
    glLineWidth((status & active_bit) ? active_link_width:
		/* normal link */       normal_link_width);

    NodeInfo& from = ep->from()->info();
    NodeInfo& to   = ep->to()->info();
    glBegin(GL_LINES);
    v3f(from.x, from.y, from.z);
    v3f(to.x,   to.y,   to.z);
    glEnd();
}


void
WebViewGL::draw_links_suffix()
{
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
}


void
WebViewGL::draw_board()
{
    glPushMatrix();

    Color3fv(strong_text_color);
    glLineWidth(normal_text_width);
    StrokeText( 0.20, -0.10, z_board(),  0.0, 0.0004,
	       web()->mapper()->x_title());
    StrokeText(-0.05,  0.20, z_board(), 90.0, 0.0004,
	       web()->mapper()->y_title());

    float  pos;
    string label;
    int i;
    for (i = 0; web()->mapper()->x_label(i, pos, label); i++) {
	glLineWidth(1);
	Color3fv(board0_color);
	glBegin(GL_LINES);
	v3f(pos, 1.00, z_board());
	v3f(pos, 1.05, z_board());
	glEnd();

	glLineWidth(normal_text_width);
	Color3fv(normal_text_color);
	StrokeText(pos, 1.05, z_board(), 40.0, 0.0003, label);
    }
    for (i = 0; web()->mapper()->y_label(i, pos, label); i++) {
	glLineWidth(1);
	Color3fv(board0_color);
	glBegin(GL_LINES);
	v3f(1.00, pos, z_board());
	v3f(1.05, pos, z_board());
	glEnd();

	glLineWidth(normal_text_width);
	Color3fv(normal_text_color);
	StrokeText(1.05, pos, z_board(), 40.0, 0.0003, label);
    }

    GLfloat v1[3], v2[3], v3[3], v4[3];
    v1[0] = 0.0;  v1[1] = 0.0;
    v2[0] = 1.0;  v2[1] = 0.0;
    v3[0] = 1.0;  v3[1] = 1.0;
    v4[0] = 0.0;  v4[1] = 1.0;
    v1[2] = v2[2] = v3[2] = v4[2] = z_board();

    Color3fv(board0_color);
    glBegin(GL_LINE_LOOP);  // glRectfv is faster?
    v3fv(v1);
    v3fv(v2);
    v3fv(v3);
    v3fv(v4);
    glEnd();
    
    n3f(0.0, 0.0, 1.0);
    if (render_mode() >= surface) {
	glBegin(GL_QUADS);
	Color3fv(board1_color);  v3fv(v1);
	Color3fv(board2_color);  v3fv(v2);
	Color3fv(board3_color);  v3fv(v3);
	Color3fv(board4_color);  v3fv(v4);
	glEnd();
    }

    glPopMatrix();
}


void
WebViewGL::draw_cursor()
{
    if (!web()->active_node()) return;
    GLfloat x = web()->active_node()->info().x;
    GLfloat y = web()->active_node()->info().y;
    GLfloat z = z_board();
    Color3fv(cursor_color);
    glLineWidth(cursor_width);
    glLoadName(cursor);
    glBegin(GL_LINES);
    v3f(x, y, 10.0);
    v3f(x, y, z);
    glEnd();
#if 0
    glBegin(GL_LINE_STRIP);
    v3f(0.0, y, 0.0);
    v3f(x,   y, 0.0);
    v3f(x, 0.0, 0.0);
    glEnd();
#endif
    glLoadName(invalid_name);
}


void
WebViewGL::pick(
    int  x,
    int  y,
    vector<int>& sysobjs,
    Web::node_vector& nodes,
    Web::link_vector& links)
{
    GLuint sel_buf[SEL_BUF_SIZE];

    strstream msg;
    msg << "(" << x << ", " << y << ") selected." << ends;
    message(msg.str());
    msg.freeze(false);
    
    make_current();
    glSelectBuffer(sizeof(sel_buf), sel_buf);
    (void) glRenderMode(GL_SELECT);
    GLint viewport[4];
    // glGetIntegerv(GL_VIEWPORT, viewport);
    viewport[0] = viewport[1] = 0;
    viewport[2] = width();
    viewport[3] = height();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPickMatrix(x, y, 5, 5, viewport);
    projection_matrix();
    draw();
    glFlush();

    GLint nhit = glRenderMode(GL_RENDER);
    DEBUGOUT("# of hits = " << nhit);
    priority_queue<vector<Hit>, less<Hit> > hits;

    GLuint* name_ptr = sel_buf;
    for (GLint i = 0; i < nhit; i++) {
	GLuint nname = *name_ptr++;
	GLuint z     = *name_ptr++;
	/* z2 */        name_ptr++;
	for (GLuint j = 0; j < nname; j++) {
	    GLuint n = *name_ptr++;
	    if (n == invalid_name) continue;
	    Hit h;
	    h.z = z;
	    h.name = n;
	    hits.push(h);
	}
    }
    
    for (Hit h; !hits.empty(); hits.pop()) {
	h = hits.top();
	if (h.name == invalid_name) continue;
	if (h.name & sysobj_bit) {
	    int sysobj_no = h.name ^ sysobj_bit;
	    DEBUGOUT("z=" << h.z << ": sysobj " << sysobj_no << " picked.");
	    sysobjs.push_back(sysobj_no);
	    continue;
	}
	if (h.name & node_bit) {
	    int node_no = h.name ^ node_bit;
	    DEBUGOUT("z=" << h.z << ": node " << node_no << " picked.");
	    nodes.push_back(node_table_[node_no]);
	    continue;
	}
	if (h.name & link_bit) {
	    int link_no = h.name ^ link_bit;
	    DEBUGOUT("z=" << h.z << ": link " << link_no << " picked.");
	    links.push_back(link_table_[link_no]);
	    continue;
	}
	DEBUGOUT("z=" << h.z << ": unknown object " << h.name << " picked.");
    }
    return;
}


void
WebViewGL::get_raster_pos(
    Web::node* vp,
    float&     x,
    float&     y)
{
    if (!vp) return;
    make_current();
    glRasterPos3f(vp->info().x, vp->info().y, vp->info().z);
    GLfloat pos[4];
    glGetFloatv(GL_CURRENT_RASTER_POSITION, pos);
    x = pos[0];
    y = pos[1];
}


void
WebViewGL::viewable_nodes(
    Web::node_vector& nodes)

{
    GLfloat matrix[16];
    make_current();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    modelview_matrix();
    glGetFloatv(GL_MODELVIEW_MATRIX, matrix);
    GLdouble modelview[16];
    int i;
    for (i = 0; i < 16; i++) modelview[i] = matrix[i];

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    projection_matrix();
    glGetFloatv(GL_PROJECTION_MATRIX, matrix);
    GLdouble projection[16];
    for (i = 0; i < 16; i++) projection[i] = matrix[i];

    GLint viewport[4];
    viewport[0] = viewport[1] = 0;
    viewport[2] = width();
    viewport[3] = height();
    
    for (Web::node_iterator vi = web()->nodes().begin();
	 vi != web()->nodes().end(); vi++) {
	if (!(*vi)->info().drawn) continue;
	GLdouble x = 0, y = 0, z = 0;
	GLint s = gluProject((*vi)->info().x, (*vi)->info().y, (*vi)->info().z,
			     modelview, projection, viewport, &x, &y, &z);
	if (s && 0 <= x && x <= width() && 0 <= y && y <= height())
	    nodes.push_back(*vi);
    }
    return;
}
