/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#include "linkview.h"

#include <fstream>
#include <strstream>
#include <X11/Xmu/Editres.h>
#include <Xm/XmAll.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include "copy_line.h"
#include "glex.h"
#include "NodeMapperUU.h"
#include "WebViewX.h"
#include "WebFrameX.h"
#include "WebExtenderHT.h"
#include "EventManagerHTX.h"
#include "Preference.h"

static void xt_control_fifo(istream*, int*, XtInputId*);
static void xt_quitbox_ok(Widget, XtPointer, XtPointer);
static void xt_toplevel_destroy(Widget, XtPointer, XtPointer);

static void hack_options(int, char*[]);
static void analyze_options(int, char*[]);
static void cleanup(void);
static int  x_error(Display*, XErrorEvent*);
static void core_dump(void);

static XtAppContext app_context   = 0;
static Widget       toplevel      = NULL;
static Widget       quitbox       = NULL;
static WebFrameX*   current_frame = NULL;
static char*        fifo_name     = "";
static XtInputId    fifo_id       = 0;

#if 0
static void xt_action_pick(Widget, XEvent*, String[], Cardinal*);
static void xt_action_drag(Widget, XEvent*, String[], Cardinal*);
#endif
static void xt_action_lift_up(Widget, XEvent*, String[], Cardinal*);
static void xt_action_lift_down(Widget, XEvent*, String[], Cardinal*);
static void xt_action_lift_down_all(Widget, XEvent*, String[], Cardinal*);

static XtActionsRec action_table[] = {
    { "lift-up",        xt_action_lift_up },
    { "lift-down",      xt_action_lift_down },
    { "lift-down-all",  xt_action_lift_down_all },
#if 0
    { "translate",      xt_action_translate },
    { "rotate",         xt_action_rotate },
    { "zoom",           xt_action_zoom },
#endif
};

// Netscape remote control
const char* progname = NULL;
const char* expected_mozilla_version = "1.1";


static void
xt_action_lift_up(Widget, XEvent*, String[], Cardinal*)
{
    current_frame->lift_up();
}


static void
xt_action_lift_down(Widget, XEvent*, String[], Cardinal*)
{
    current_frame->lift_down();
}


static void
xt_action_lift_down_all(Widget, XEvent*, String[], Cardinal*)
{
    current_frame->lift_down_all();
}


static void 
xt_control_fifo(
    istream* in,
    int*, XtInputId*)
{
    strstream buf;
    string cmd;
    string arg, arg2;
    int    para = 0;
    int    ch   = '\0';

    WebView* view = current_frame->active_view();

    copy_line(*in, buf); buf << ends;
    if (!*in) {
	current_frame->message("ERROR: Control FIFO broken.");
	XtRemoveInput(fifo_id);
	return;
    }
    DEBUGOUT(buf.str());
    buf.freeze(false);

    buf >> cmd;
    if (cmd.length() == 0) goto done;
    ch = cmd[0];

    switch (ch) {
	/* Home */     case 'H':  view->home_view();     goto done;
	/* Up */       case 'U':  view->lift_up();       goto done;
	/* Down */     case 'D':  view->lift_down();     goto done;
	/* Down All */ case 'A':  view->lift_down_all(); goto done;
	/* Browse */   case 'B':  view->browse();        goto done;
	/* Reload */   case 'L':  view->reload();        goto done;
	/* Quit */     case 'Q':  Quit();                goto done;
	/* Exit */     case 'X':  Quit();                goto done;
    }

    buf >> arg;
    switch (ch) {
     case 'C': /* CURSOR */
	if (arg.length()) view->focus(arg);
	goto done;

     case 'M': /* MARK */
	buf >> arg2;
	if (arg.length() && arg2.length()) view->set_mark(arg, arg2);
	else if (arg.length())             view->set_mark(arg);
	else                               view->set_mark();
	goto done;
	
     case 'T':
	if (!cmd[1] || !cmd[2]) break;
	para = 0;
	if (arg.length()) para = atoi(arg.c_str());

	switch ((toupper(cmd[1]) << 8) | cmd[2]) {
	 case ('X' << 8) | '+':
	    if (para) view->trans_left(para);
	    else      view->trans_left();
	    break;

	 case ('X' << 8) | '-':
	    if (para) view->trans_right(para);
	    else      view->trans_right();
	    break;

	 case ('Y' << 8) | '+':
	    if (para) view->trans_up(para);
	    else      view->trans_up();
	    break;

	 case ('Y' << 8) | '-':
	    if (para) view->trans_down(para);
	    else      view->trans_down();
	    break;
	}
	goto done;

     case 'R':
	if (!cmd[1] || !cmd[2]) break;
	para = 0;
	if (arg.length()) para = atoi(arg.c_str());
	switch ((toupper(cmd[1]) << 8) | cmd[2]) {

	 case ('X' << 8) | '+':
	    if (para) view->rot_up(para);
	    else      view->rot_up();
	    break;

	 case ('X' << 8) | '-':
	    if (para) view->rot_down(para);
	    else      view->rot_down();
	    break;
	 
	 case ('Y' << 8) | '+':
	    if (para) view->rot_left(para);
	    else      view->rot_left();
	    break;
	 
	 case ('Y' << 8) | '-':
	    if (para) view->rot_right(para);
	    else      view->rot_right();
	    break;
	}
	goto done;

     case 'Z':
	if (!cmd[1] || !cmd[2]) break;
	para = 0;
	if (arg.length()) para = atoi(arg.c_str());

	switch ((toupper(cmd[1]) << 8) | cmd[2]) {
	 case ('Z' << 8) | '+':
	    if (para) view->zoom_in(para);
	    else      view->zoom_in();
	    break;

	 case ('Z' << 8) | '-':
	    if (para) view->zoom_out(para);
	    else      view->zoom_out();
	    break;
	}
	goto done;

    }

    return;  /* no match */

  done:;  /* OK! */
    view->update();
}


int
main(
    int   argc,
    char* argv[])
{
    atexit(cleanup);
    // set_terminate(core_dump);
    
    progname= strchr(argv[0], '/');
    if (progname) progname++;
    else	  progname = argv[0];

    // help and version options (before opening display)
    hack_options(argc, argv);

    // Motif
    XtToolkitInitialize();
    XtSetLanguageProc(NULL, NULL, NULL);

    toplevel = XtAppInitialize(&app_context, "Nattoview", NULL, 0,
			       &argc, argv, NULL, NULL, 0);
    XtAddCallback(toplevel, XmNdestroyCallback, xt_toplevel_destroy, NULL);
    XtAppAddActions(app_context, action_table, XtNumber(action_table));

    // other options
    analyze_options(argc, argv);

    // debugging X (when not static linked)
    if (preference.debug_X) {
	XSynchronize(XtDisplay(toplevel), True);
	XSetErrorHandler(x_error);
    }

    char* server  = getenv("NATTOVIEW_LINKDB");
    Web*  web     = new Web(new NodeMapperUU);
    current_frame = new WebFrameX(web,
				  server ? server: LINKDB_HOST, LINKDB_PORT,
				  toplevel);
    XtRealizeWidget(toplevel);

    // initialize W3C libwww
    HTLibInit("NattoView", NATTOVIEW);
    if (preference.debug_libwww.length()) 
	HTSetTraceMessageMask(preference.debug_libwww.c_str());
    (void) new EventManagerHTX(app_context);  // singleton
    
    // open control FIFO
    fifo_name = getenv("NATTOVIEW_FIFO");
    if (fifo_name) {
	// strstream msg;
	// cerr << "Opening control FIFO " << fifo_name << "..." << endl;
	// msg.freeze(false);
	ifstream* fifo = new ifstream(fifo_name, ios::in);
	if (!fifo || !*fifo) {
	    strstream msg;
	    msg << "Warning: Can't open control FIFO "
		<< fifo_name << '.' << ends;
	    current_frame->message(msg.str());
	    msg.freeze(false);
	}
	else {
	    fifo_id = XtAppAddInput(app_context, fifo->rdbuf()->fd(),
				    (XtPointer) XtInputReadMask,
				    (XtInputCallbackProc) xt_control_fifo,
				    (XtPointer) fifo);
	}
    }
    
    // initial access
    current_frame->focus(preference.home_page ?
			 preference.home_page: URL(HOME_PAGE));

    // make "Realy Quit ?" dialog
    quitbox = XmCreateWarningDialog(toplevel, "quitbox", NULL, 0);
    XtUnmanageChild(XmMessageBoxGetChild(quitbox,
					 XmDIALOG_HELP_BUTTON));
    XtAddCallback(quitbox, XmNokCallback, xt_quitbox_ok, NULL);

    // set signal handler
    signal(SIGPIPE, SIG_IGN);
    
    // event loop
    XtAddEventHandler(toplevel, 0, True, _XEditResCheckMessages, NULL);
    for (;;) {
	const XtInputMask no_input = XtIMAll & ~XtIMAlternateInput;
	while (XtAppPending(app_context) & no_input)
	    XtAppProcessEvent(app_context, no_input);
	XtAppProcessEvent(app_context, XtIMAll);
    }
    return EXIT_SUCCESS;  // dummy
}


static void
hack_options(
    int   argc,
    char* argv[])
{
    for (int i = 1; i < argc; i++) {
	if (strncmp(argv[i], "-v",  2) == 0 || strncmp(argv[i], "--v", 3) == 0) {
	    cout << "The Natto View version " NATTOVIEW "\n"
		    "Copyright (c) 1995-1997  Hidekazu Shiozawa,\n"
		    "                         Keio University (Matsushita Lab.)\n\n"
		    "Bug reports are wellcome!! <shiozawa@myo.inst.keio.ac.jp>\n" << endl;
	    exit(EXIT_SUCCESS);
	}
	if (strncmp(argv[i], "-h",  2) == 0 || strncmp(argv[i], "--h", 3) == 0 ||
	    strncmp(argv[i], "-?",  2) == 0 || strncmp(argv[i], "--?", 3) == 0) {
	    cout << "The Natto View version " NATTOVIEW "\n"
		    "Usage: " << progname << " [-options...] [-toolkit-options...] [URL]\n\n"
		    "Options:\n"
		    "  -v, -version, --version  print version information\n"
		    "  -h, -help,    --help     print this message\n"
		    "  -install                 install colormap\n"
		 << endl;
	    exit(EXIT_SUCCESS);
	}
    }
}


static void
analyze_options(
    int   argc,
    char* argv[])
{
    for (int i = 1; i < argc && argv[i]; i++) {
	if (argv[i][0] != '-') {
	    preference.home_page = argv[i];
	}
	else if (strcmp(argv[i], "-install") == 0) {
	    preference.install_cmap = true;
	}
	else if (strncmp(argv[i], "-wwwdebug=", 10) == 0) {
	    preference.debug_libwww = argv[i] + 10;
	}
	else if (strcmp(argv[i], "-xdebug") == 0) {
	    preference.debug_X = true;
	}
	else if (strncmp(argv[i], "-proxy=", 7) == 0) {
	    preference.http_proxy = argv[i] + 7;
	}
	else if (strncmp(argv[i], "-ftpproxy=", 10) == 0) {
	    preference.ftp_proxy = argv[i] + 10;
	}
	else {
	    cout << "Illegal option: " << argv[i] << endl
		<< argv[0] << " -h for help." << endl;
	    exit(EXIT_SUCCESS);
	}
    }
}


static void
xt_toplevel_destroy(Widget, XtPointer, XtPointer)
{
    exit(EXIT_SUCCESS);
}


void 
Quit(void)
{
    XtManageChild(quitbox);
}


static void
xt_quitbox_ok(Widget, XtPointer, XtPointer)
{
    exit(EXIT_SUCCESS);
}


static void
cleanup(void)
{
    // HTConversion_deleteAll(converters);
    HTLibTerminate();
}


// catch X error
static int
x_error(Display*, XErrorEvent*)
{
    core_dump();
    return 1;
}


// Dump core! in order to stack trace
static void
core_dump(void)
{
    // Segmemtation violation on purpose. 
    char* p = 0;
    *p = 0;
}
