/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996-1997  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//                           Keio University (Matsushita Lab.)
//                           All rights reserved.
//  Please first read the full copyright statement in the file LICENSE.
//  Permission to copy this software, to redistribute it on a nonprofit
//  basis, and to use it for any purpose, is hereby granted, provided that 
//  this license information and copyright notice appear in all copies. 
*/

#ifndef WWW_H
#define WWW_H
#ifdef __GNUG__
#pragma interface
#endif

//////////////////////////////////////////////////////////////////////////////
//  www.h -- libwww interface
//

#include <string>
#include <istream>
#include <ostream>
#include <strstream>
#include <time.h>
#include <stdio.h>
#include <sys/types.h>
#include <WWWLib.h>
#include <WWWHTTP.h>
#include <WWWHTML.h>
#include <WWWMIME.h>
#include <WWWUtil.h>


//////////////////////////////////////////////////////////////////////////////
//  URL -- canonicalization and string conversion of URL(URI)
//
class URL
{
    friend istream& operator>>(istream&, URL&);
 public:
    // primitive functions
    URL() {}
    URL(const URL& url):          absolute_(url.absolute_) {}
    URL(string url):              absolute_(htparse(url)) {}
    URL(string url, URL base):    absolute_(htparse(url, base.absolute())) {}
    URL(string url, string base): absolute_(htparse(url, base)) {}
    URL(HTAnchor* htanchor);
    virtual ~URL() {}
    
    URL& operator=(URL url)
	{ absolute_ = url.absolute_; return *this; }
    URL& operator=(string url)
	{ absolute_ = htparse(url, this->absolute_); return *this; }
    
    operator bool() { return absolute_.length(); }

    // convert to string
    string absolute()            const { return absolute_; }
    string relative(string base) const;
    string relative(URL base)    const { return relative(base.absolute_); }

    // extract the last meaingful word, mostly file name
    string last_word() const;

    // for parameter of parse()
    enum field_t {
	method     = PARSE_ACCESS,
	host       = PARSE_HOST,
	path       = PARSE_PATH,
	frag       = PARSE_ANCHOR,
	punct      = PARSE_PUNCTUATION,
	all_fields = PARSE_ALL
    };
    
    // split and extract URL fields
    string parse(int fields);
    
 protected:
    // HTParse() wrapper
    static string htparse(string rel, string base);
    static string htparse(string rel);

 private:
    string absolute_;  // absolute URL
};

// comparison as string
inline bool operator<(URL url1, URL url2)
    { return url1.absolute() < url2.absolute(); }

// stream I/O
inline ostream& operator<<(ostream& out, URL url)
    { return out << url.absolute(); }

inline istream& operator>>(istream& in, URL& url)
    { string s;  in >> s;  url = s;  return in; }


//////////////////////////////////////////////////////////////////////////////
// SGMLParser -- libwww wrapper for parsing SGML tags
//
class SGMLParser;

// libwww interface
struct _HTStream {
    const HTStreamClass* isa;
    HTStream*            target;
    HTRequest*           request;
    SOCKET               sockfd;
    int                  version;
    int                  state;
    char*                url;
    BOOL                 transparent;
};

struct _HTStructured {
    const HTStructuredClass* isa;
    SGMLParser*              wrapper;
};


class SGMLParser
{
 protected:
    // libwww interface
    /*const*/ string  class_name_;
    HTStructured*     me_;
    HTStream*         htstream_;
    HTStructuredClass structured_class_;
    HTRequest*        htrequest_;
    
 public:
    // primitive functions
    SGMLParser(string name, const SGML_dtd& dtd, HTRequest* req);
    virtual ~SGMLParser();

    // member access
    HTStream*  htstream()  { return htstream_; }
    HTRequest* htrequest() { return htrequest_; }
    URL        current_url() const;

    // stream callbacks
    virtual int  flush() { return HT_OK; }
    virtual int  free();
    virtual int  abort(HTList* errors);

    virtual int  put_character(char)
	{ return HT_OK; }
    virtual int  put_string(const char* str)
	{ while (*str) put_character(*str++);  return HT_OK; }
    virtual int  put_block(const char* block, int len)
	{ while (len-- > 0) put_character(*block++);  return HT_OK; }

    virtual void start_element(int /*element*/ , const char* /*present*/,
			       const char** /*value*/) {}
    virtual void end_element(int /*element*/) {}
    virtual void put_entity(int /*entity_number*/) {}

 protected:
    // libwww callback interface
    static int  ht_flush(HTStructured* me)
	{ return me->wrapper->flush(); }
    static int  ht_free(HTStructured* me)
	{ return me->wrapper->free(); }
    static int  ht_abort(HTStructured* me, HTList* errors)
	{ return me->wrapper->abort(errors); }
    static int  ht_put_character(HTStructured* me, char c)
	{ return me->wrapper->put_character(c); }
    static int  ht_put_string(HTStructured* me, const char* str)
	{ return me->wrapper->put_string(str); }
    static int  ht_put_block(HTStructured* me, const char* block, int len)
	{ return me->wrapper->put_block(block, len); }
    static void ht_start_element(HTStructured* me, int element,
				const char* present, const char** value)
	{ me->wrapper->start_element(element, present, value); }
    static void ht_end_element(HTStructured* me, int element)
	{ me->wrapper->end_element(element); }
    static void ht_put_entity(HTStructured* me, int entity_number)
	{ me->wrapper->put_entity(entity_number); }
};


//////////////////////////////////////////////////////////////////////////////
//  HTMLParser -- parsing HTML tags
//
class HTMLParser: public SGMLParser
{
 public:
    // primitive functions
    HTMLParser(HTRequest*);
    virtual ~HTMLParser() {}

    // reset the state of parser
    void reset();

    // tag callbacks
    enum heading_t { H_none, H1, H2, H3, H4, H5, H6, H7, TITLE };
    virtual void title(string) {}
    virtual void heading(heading_t, string) {}
    virtual void hyperlink(URL, string /*anchor*/, 
			   string /*title*/, string /*type*/) {}
    virtual void image(URL, string /* alt */ = "") {}
    virtual void movie(URL, string /* alt */ = "") {}
    virtual void button(URL, string /*method*/) {}
    virtual void frame(URL) {}
    virtual void object(URL) {}

    // stream callbacks (overrided)
    virtual int  put_character(char);
    virtual void start_element(int, const char*, const char**);
    virtual void end_element(int);

 protected:
    // make URL with current context
    URL url(string relative);
    string content();
    static string trim_space(string str);
    
 private:
    bool      in_header_;  // in <HEAD> ?
    bool      isindex_;    // <ISINDEX> ?
    heading_t in_heading_; // in <Hn> ?
    bool      in_anchor_;  // in <A> ?
    char      prev_char_;  // previous character
    string    h_str_;      // <Hn>...</Hn>
    string    anchor_;     // <A>...<A>
    string    href_url_;   // <A HREF="...">
    string    href_title_; // <A TITLE="...">
    string    href_rel_;   // <A REL="...">
    URL       base_;       // <BASE HREF="...">
    strstream content_;
};

#endif  /* WWW_H */
